/*
 * Copyright 2006, 2007 Odysseus Software GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.odysseus.el.misc;

import java.math.BigDecimal;
import java.math.BigInteger;

import javax.el.ELException;

/**
 * Arithmetic Operations as specified in chapter 1.7.
 * 
 * @author Christoph Beck
 */
public class NumberOperations {
	private final static Long LONG_ZERO = Long.valueOf(0L);

	private final static boolean isDotEe(String value) {
		int length = value.length();
		for (int i = 0; i < length; i++) {
			switch (value.charAt(i)) {
				case '.':
				case 'E':
				case 'e': return true;
			}
		}
		return false;
	}

	private final static boolean isDotEe(Object value) {
		return value instanceof String && isDotEe((String)value);
	}

	private static final boolean isFloatOrDouble(Object value) {
		return value instanceof Float || value instanceof Double;
	}

	private static final boolean isFloatOrDoubleOrDotEe(Object value) {
		return isFloatOrDouble(value) || isDotEe(value);
	}

	private static final boolean isBigDecimalOrBigInteger(Object value) {
		return value instanceof BigDecimal || value instanceof BigInteger;
	}

	private static final boolean isBigDecimalOrFloatOrDoubleOrDotEe(Object value) {
		return value instanceof BigDecimal || isFloatOrDoubleOrDotEe(value);
	}

	private static final BigDecimal toBigDecimal(Object value) {
		return TypeConversions.coerceToNumber(value, BigDecimal.class);
	}

	private static final BigInteger toBigInteger(Object value) {
		return TypeConversions.coerceToNumber(value, BigInteger.class);
	}

	private static final Double toDouble(Object value) {
		return TypeConversions.coerceToNumber(value, Double.class);
	}

	private static final Long toLong(Object value) {
		return TypeConversions.coerceToNumber(value, Long.class);
	}

	public static final Number add(Object o1, Object o2) {
		if (o1 == null && o2 == null) {
			return LONG_ZERO;
		}
		if (o1 instanceof BigDecimal || o2 instanceof BigDecimal) {
			return toBigDecimal(o1).add(toBigDecimal(o2));
		}
		if (isFloatOrDoubleOrDotEe(o1) || isFloatOrDoubleOrDotEe(o2)) {
			if (o1 instanceof BigInteger || o2 instanceof BigInteger) {
				return toBigDecimal(o1).add(toBigDecimal(o2));
			}
			return toDouble(o1) + toDouble(o2);
		}
		if (o1 instanceof BigInteger || o2 instanceof BigInteger) {
			return toBigInteger(o1).add(toBigInteger(o2));
		}
		return toLong(o1) + toLong(o2);
	}

	public static final Number sub(Object o1, Object o2) {
		if (o1 == null && o2 == null) {
			return LONG_ZERO;
		}
		if (o1 instanceof BigDecimal || o2 instanceof BigDecimal) {
			return toBigDecimal(o1).subtract(toBigDecimal(o2));
		}
		if (isFloatOrDoubleOrDotEe(o1) || isFloatOrDoubleOrDotEe(o2)) {
			if (o1 instanceof BigInteger || o2 instanceof BigInteger) {
				return toBigDecimal(o1).subtract(toBigDecimal(o2));
			}
			return toDouble(o1) - toDouble(o2);
		}
		if (o1 instanceof BigInteger || o2 instanceof BigInteger) {
			return toBigInteger(o1).subtract(toBigInteger(o2));
		}
		return toLong(o1) - toLong(o2);
	}

	public static final Number mul(Object o1, Object o2) {
		if (o1 == null && o2 == null) {
			return LONG_ZERO;
		}
		if (o1 instanceof BigDecimal || o2 instanceof BigDecimal) {
			return toBigDecimal(o1).multiply(toBigDecimal(o2));
		}
		if (isFloatOrDoubleOrDotEe(o1) || isFloatOrDoubleOrDotEe(o2)) {
			if (o1 instanceof BigInteger || o2 instanceof BigInteger) {
				return toBigDecimal(o1).multiply(toBigDecimal(o2));
			}
			return toDouble(o1) * toDouble(o2);
		}
		if (o1 instanceof BigInteger || o2 instanceof BigInteger) {
			return toBigInteger(o1).multiply(toBigInteger(o2));
		}
		return toLong(o1) * toLong(o2);
	}

	public static final Number div(Object o1, Object o2) {
		if (o1 == null && o2 == null) {
			return LONG_ZERO;
		}
		if (isBigDecimalOrBigInteger(o1) || isBigDecimalOrBigInteger(o2)) {
			return toBigDecimal(o1).divide(toBigDecimal(o2), BigDecimal.ROUND_HALF_UP);
		}
		return toDouble(o1) / toDouble(o2);
	}

	public static final Number mod(Object o1, Object o2) {
		if (o1 == null && o2 == null) {
			return LONG_ZERO;
		}
		if (isBigDecimalOrFloatOrDoubleOrDotEe(o1) || isBigDecimalOrFloatOrDoubleOrDotEe(o2)) {
			return toDouble(o1) % toDouble(o2);
		}
		if (o1 instanceof BigInteger || o2 instanceof BigInteger) {
			return toBigInteger(o1).remainder(toBigInteger(o2));
		}
		return toLong(o1) % toLong(o2);
	}

	public static final Number neg(Object value) {
		if (value == null) {
			return LONG_ZERO;
		}
		if (value instanceof BigDecimal) {
			return ((BigDecimal)value).negate();
		}
		if (value instanceof BigInteger) {
			return ((BigInteger)value).negate();
		}
		if (value instanceof Double) {
			return Double.valueOf(-((Double)value).doubleValue());
		}
		if (value instanceof Float) {
			return Float.valueOf(-((Float)value).floatValue());
		}
		if (value instanceof String) {
			if (isDotEe((String)value)) {
				return Double.valueOf(-toDouble(value).doubleValue());
			}
			return Long.valueOf(-toLong(value).longValue());
		}
		if (value instanceof Long) {
			return Long.valueOf(-((Long)value).longValue());
		}
		if (value instanceof Integer) {
			return Integer.valueOf(-((Integer)value).intValue());
		}
		if (value instanceof Short) {
			return Short.valueOf((short)-((Short)value).shortValue());
		}
		if (value instanceof Byte) {
			return Byte.valueOf((byte)-((Byte)value).byteValue());
		}
		throw new ELException(LocalMessages.get("error.negate", value.getClass()));
	}
}
