/*
 * Copyright 2006, 2007 Odysseus Software GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.odysseus.el.tree.impl;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

import de.odysseus.el.misc.LocalMessages;
import de.odysseus.el.tree.FunctionNode;
import de.odysseus.el.tree.IdentifierNode;
import de.odysseus.el.tree.Tree;

import de.odysseus.el.tree.impl.Scanner.ScanException;
import de.odysseus.el.tree.impl.Scanner.Symbol;
import de.odysseus.el.tree.impl.ast.*;

import static de.odysseus.el.tree.impl.Scanner.Symbol.*;

/**
 * Handcrafted top-down parser.
 *
 * @author Christoph Beck
 */
final class Parser {
	/**
	 * Parse exception type
	 */
	@SuppressWarnings("serial")
	static class ParseException extends Exception {
		ParseException(int position, String encountered, String expected) {
			super(LocalMessages.get("error.parse", position, encountered, expected));
		}
	}

	/**
	 * Token type (used to store lookahead)
	 */
	private static final class LookaheadToken {
		final Symbol symbol;
		final String image;
		final int position;

		LookaheadToken(Symbol symbol, String image, int position) {
			this.symbol = symbol;
			this.image = image;
			this.position = position;
		}
	}

	private static final String EXPR_FIRST =
		IDENTIFIER + "|" + 
		STRING + "|" + FLOAT + "|" + INTEGER + "|" + TRUE + "|" + FALSE + "|" + NULL + "|" +
		MINUS + "|" + NOT + "|" + EMPTY + "|" +
		LPAREN;

	private final Builder context;
	private final Scanner scanner;

	private List<IdentifierNode> identifiers = Collections.emptyList();
	private List<FunctionNode> functions = Collections.emptyList();
	private List<LookaheadToken> lookahead = Collections.emptyList();

	public Parser(Builder context, String input) {
		this.context = context;
		this.scanner = new Scanner(input);
	}

	private Symbol symbol; // current token symbol
	private String image; // current token text
	private int position; // current token position

	private Number parseInteger(String string) throws ParseException {
		try {
			return context.parseInteger(string);
		} catch (NumberFormatException e) {
			fail(INTEGER);
			return null;
		}
	}
	
	private Number parseFloat(String string) throws ParseException {
		try {
			return context.parseFloat(string);
		} catch (NumberFormatException e) {
			fail(FLOAT);
			return null;
		}
	}

	/**
	 * throw exception
	 */
	private void fail(String expected) throws ParseException {
		String encountered = image == null ? symbol.toString() : "'" + image + "'";
		throw new ParseException(position, encountered, expected);
	}

	/**
	 * throw exception
	 */
	private void fail(Symbol expected) throws ParseException {
		fail(expected.toString());
	}

	/**
	 * get lookahead symbol.
	 */
	private Symbol lookahead(int index) throws ScanException, ParseException {
		if (lookahead.isEmpty()) {
			lookahead = new LinkedList<LookaheadToken>();
		}
		while (index >= lookahead.size()) {
			lookahead.add(new LookaheadToken(scanner.next(), scanner.getImage(), scanner.getPosition()));
		}
		return lookahead.get(index).symbol;
	}

	/**
	 * consume current token (get next token)
	 */
	private void consumeToken() throws ScanException, ParseException {
		if (lookahead.isEmpty()) {
			symbol = scanner.next();
			image = scanner.getImage();
			position = scanner.getPosition();
		} else {
			LookaheadToken token = lookahead.remove(0);
			symbol = token.symbol;
			image = token.image;
			position = token.position;
		}
	}

	/**
	 * consume current token (get next token); throw exception if the current token doesn't
	 * match the expected symbol.
	 */
	private void consumeToken(Symbol expected) throws ScanException, ParseException {
		if (symbol != expected) {
			fail(expected);
		}
		consumeToken();
	}
	
	/**
	 * tree := text? ((dynamic text?)+ | (deferred text?)+)? 
	 */
	public Tree tree() throws ScanException, ParseException {
		consumeToken();
		AstNode t = text();
		if (symbol == EOF) {
			if (t == null) {
				t = new AstText("");
			}
			return new Tree(t, functions, identifiers, false);
		}
		AstEval e = eval();
		if (symbol == EOF && t == null) {
			return new Tree(e, functions, identifiers, e.isDeferred());
		}
		ArrayList<AstNode> list = new ArrayList<AstNode>();
		if (t != null) {
			list.add(t);
		}
		list.add(e);
		t = text();
		if (t != null) {
			list.add(t);
		}
		while (symbol != EOF) {
			if (e.isDeferred()) {
				list.add(eval(true, true));
			} else {
				list.add(eval(true, false));
			}
			t = text();
			if (t != null) {
				list.add(t);
			}
		}
		return new Tree(new AstComposite(list), functions, identifiers, e.isDeferred());
	}

	/**
	 * text := <TEXT>
	 */
	private AstNode text() throws ScanException, ParseException {
		AstNode v = null;
		if (symbol == TEXT) {
			v = new AstText(image);
			consumeToken();
		}
		return v;
	}

	/**
	 * eval := dynamic | deferred
	 */
	private AstEval eval() throws ScanException, ParseException {
		AstEval e = eval(false, false);
		if (e == null) {
			e = eval(false, true);
			if (e == null) {
				fail(START_EVAL_DEFERRED + "|" + START_EVAL_DYNAMIC);
			}
		}
		return e;
	}

	/**
	 * dynmamic := <START_EVAL_DYNAMIC> expr <END_EVAL>
	 * deferred := <START_EVAL_DEFERRED> expr <END_EVAL>
	 */
	private AstEval eval(boolean required, boolean deferred) throws ScanException, ParseException {
		AstEval v = null;
		Symbol start_eval = deferred ? START_EVAL_DEFERRED : START_EVAL_DYNAMIC;
		if (symbol == start_eval) {
			consumeToken();
			v = new AstEval(expr(true), deferred);
			consumeToken(END_EVAL);
		} else if (required) {
			fail(start_eval);
		}
		return v;
	}

	/**
	 * expr := or (<QUESTION> expr <COLON> expr)?
	 */
	private AstNode expr(boolean required) throws ScanException, ParseException {
		AstNode v = or(required);
		if (v == null) {
			return null;
		}
		if (symbol == QUESTION) {
			consumeToken();
			AstNode a = expr(true);
			consumeToken(COLON);
			AstNode b = expr(true);
			v = new AstChoice(v, a, b);
		}
		return v;
	}

	/**
	 * or := and (<OR> and)*
	 */
	private AstNode or(boolean required) throws ScanException, ParseException {
		AstNode v = and(required);
		if (v == null) {
			return null;
		}
		while (symbol == OR) {
			consumeToken();
			v = new AstBinary(v, and(true), AstBinary.OR);
		}
		return v;
	}

	/**
	 * and := eq (<AND> eq)*
	 */
	private AstNode and(boolean required) throws ScanException, ParseException {
		AstNode v = eq(required);
		if (v == null) {
			return null;
		}
		while (symbol == AND) {
			consumeToken();
			v = new AstBinary(v, eq(true), AstBinary.AND);
		}
		return v;
	}

	/**
	 * eq := cmp (<EQ> cmp | <NE> cmp)*
	 */
	private AstNode eq(boolean required) throws ScanException, ParseException {
		AstNode v = cmp(required);
		if (v == null) {
			return null;
		}
		while (true) {
			switch (symbol) {
				case EQ:
					consumeToken();
					v = new AstBinary(v, cmp(true), AstBinary.EQ);
					break;
				case NE:
					consumeToken();
					v = new AstBinary(v, cmp(true), AstBinary.NE);
					break;
				default:
					return v;
			}
		}
	}

	/**
	 * cmp := add (<LT> add | <LE> add | <GE> add | <GT> add)*
	 */
	private AstNode cmp(boolean required) throws ScanException, ParseException {
		AstNode v = add(required);
		if (v == null) {
			return null;
		}
		while (true) {
			switch (symbol) {
				case LT:
					consumeToken();
					v = new AstBinary(v, add(true), AstBinary.LT);
					break;
				case LE:
					consumeToken();
					v = new AstBinary(v, add(true), AstBinary.LE);
					break;
				case GE:
					consumeToken();
					v = new AstBinary(v, add(true), AstBinary.GE);
					break;
				case GT:
					consumeToken();
					v = new AstBinary(v, add(true), AstBinary.GT);
					break;
				default:
					return v;
			}
		}
	}

	/**
	 * add := add (<PLUS> mul | <MINUS> mul)*
	 */
	private AstNode add(boolean required) throws ScanException, ParseException {
		AstNode v = mul(required);
		if (v == null) {
			return null;
		}
		while (true) {
			switch (symbol) {
				case PLUS:
					consumeToken();
					v = new AstBinary(v, mul(true), AstBinary.ADD);
					break;
				case MINUS:
					consumeToken();
					v = new AstBinary(v, mul(true), AstBinary.SUB);
					break;
				default:
					return v;
			}
		}
	}

	/**
	 * mul := unary (<MUL> unary | <DIV> unary | <MOD> unary)*
	 */
	private AstNode mul(boolean required) throws ScanException, ParseException {
		AstNode v = unary(required);
		if (v == null) {
			return null;
		}
		while (true) {
			switch (symbol) {
				case MUL:
					consumeToken();
					v = new AstBinary(v, unary(true), AstBinary.MUL);
					break;
				case DIV:
					consumeToken();
					v = new AstBinary(v, unary(true), AstBinary.DIV);
					break;
				case MOD:
					consumeToken();
					v = new AstBinary(v, unary(true), AstBinary.MOD);
					break;
				default:
					return v;
			}
		}
	}

	/**
	 * unary := <NOT> unary | <MINUS> unary | <EMPTY> unary | value
	 */
	private AstNode unary(boolean required) throws ScanException, ParseException {
		AstNode v = null;
		switch (symbol) {
			case NOT:
				consumeToken();
				v = new AstUnary(unary(true), AstUnary.NOT);
				break;
			case MINUS:
				consumeToken();
				v = new AstUnary(unary(true), AstUnary.NEG);
				break;
			case EMPTY:
				consumeToken();
				v = new AstUnary(unary(true), AstUnary.EMPTY);
				break;
			default:
				v = value();
		}
		if (v == null && required) {
			fail(EXPR_FIRST);
		}
		return v;
	}

	/**
	 * value := (nonliteral | literal) (<DOT> <IDENTIFIER> | <LBRACK> expr <RBRACK>)*
	 */
	private AstNode value() throws ScanException, ParseException {
		boolean lvalue = true;
		AstNode v = nonliteral();
		if (v == null) {
			v = literal();
			if (v == null) {
				return null;
			}
			lvalue = false;
		}
		while (true) {
			switch (symbol) {
				case DOT:
					consumeToken();
					String name = image;
					consumeToken(IDENTIFIER);
					if (symbol == LPAREN && context.isEnabled(Builder.Feature.METHOD_INVOCATIONS)) {
						consumeToken();
						v = new AstMethod(v, name, list());
						consumeToken(RPAREN);
					} else {
						v = new AstDot(v, name, lvalue);
					}
					break;
				case LBRACK:
					consumeToken();
					AstNode property = expr(true);
					boolean strict = !context.isEnabled(Builder.Feature.NULL_PROPERTIES);
					v = new AstBracket(v, property, lvalue, strict);
					consumeToken(RBRACK);
					break;
				default:
					return v;
			}
		}
	}

	/**
	 * nonliteral := <IDENTIFIER> | function | <LPAREN> expr <RPAREN>
	 * function   := (<IDENTIFIER> <COLON>)? <IDENTIFIER> <LPAREN> list? <RPAREN>
	 */
	private AstNode nonliteral() throws ScanException, ParseException {
		AstNode v = null;
		switch (symbol) {
			case IDENTIFIER:
				String name = image;
				consumeToken();
				if (symbol == COLON && lookahead(0) == IDENTIFIER && lookahead(1) == LPAREN) { // ns:f(...)
					consumeToken();
					name += ":" + image;
					consumeToken();
				}
				if (symbol == LPAREN) { // function
					consumeToken();
					List<AstNode> args = list();
					consumeToken(RPAREN);
					if (functions.isEmpty()) {
						functions = new ArrayList<FunctionNode>(4);
					}
					AstFunction function = new AstFunction(name, functions.size(), args);
					functions.add(function);
					v = function;
				} else { // identifier
					if (identifiers.isEmpty()) {
						identifiers = new ArrayList<IdentifierNode>(4);
					}
					AstIdentifier identifier = new AstIdentifier(name, identifiers.size());
					identifiers.add(identifier);
					v = identifier;
				}
				break;
			case LPAREN:
				consumeToken();
				v = expr(true);
				consumeToken(RPAREN);
				v = new AstNested(v);
				break;
		}
		return v;
	}

	/**
	 * list := expr (<COMMA> expr)*
	 */
	private List<AstNode> list() throws ScanException, ParseException {
		List<AstNode> l = null;
		AstNode v = expr(false);
		if (v != null) {
			l = new ArrayList<AstNode>();
			l.add(v);
			while (symbol == COMMA) {
				consumeToken();
				l.add(expr(true));
			}
		}
		return l;
	}
	
	/**
	 * literal := <TRUE> | <FALSE> | <STRING> | <INTEGER> | <FLOAT> | <NULL>
	 */
	private AstNode literal() throws ScanException, ParseException {
		AstNode v = null;
		switch (symbol) {
			case TRUE:
				v = new AstBoolean(true);
				consumeToken();
				break;
			case FALSE:
				v = new AstBoolean(false);
				consumeToken();
				break;
			case STRING:
				v = new AstString(image);
				consumeToken();
				break;
			case INTEGER:
				v = new AstNumber(parseInteger(image));
				consumeToken();
				break;
			case FLOAT:
				v = new AstNumber(parseFloat(image));
				consumeToken();
				break;			
			case NULL:
				v = new AstNull();
				consumeToken();
				break;			
		}
		return v;
	}
}
