/*
 * Copyright 2006, 2007 Odysseus Software GmbH
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.odysseus.el.tree.impl.ast;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;

import javax.el.ELContext;
import javax.el.ELException;

import de.odysseus.el.misc.LocalMessages;
import de.odysseus.el.misc.TypeConversions;
import de.odysseus.el.tree.Bindings;
import de.odysseus.el.tree.FunctionNode;

public final class AstFunction extends AstRightValue implements FunctionNode {
	private final String name;
	private final int index;
	private final List<AstNode> nodes;

	public AstFunction(String name, int index, List<AstNode> nodes) {
		this.nodes = nodes;
		this.name = name;
		this.index = index;
	}

	@Override 
	public Object eval(Bindings bindings, ELContext context) {
		Method method = bindings.getFunction(index);
		Class[] types = method.getParameterTypes();
		Object[] params = null;
		if (types.length > 0) {
			params = new Object[types.length];
			for (int i = 0; i < params.length; i++) {
				Object param = nodes.get(i).eval(bindings, context);
				if (param != null || types[i].isPrimitive()) {
					params[i] = TypeConversions.coerceToType(param, types[i]);
				}
			}
		}
		try {
			return method.invoke(null, params);
		} catch (IllegalAccessException e) {
			throw new ELException(LocalMessages.get("error.function.access", name), e);
		} catch (InvocationTargetException e) {
			throw new ELException(LocalMessages.get("error.function.invocation", name), e.getCause());
		}
	}

	@Override
	public String toString() {
		return name + "(...)";
	}	

	@Override 
	public void appendStructure(StringBuilder b, Bindings bindings) {
		b.append(bindings == null ? name : "<fn>");
		b.append("(");
		if (getCardinality() > 0) {
			nodes.get(0).appendStructure(b, bindings);
			for (int i = 1; i < getCardinality(); i++) {
				b.append(", ");
				nodes.get(i).appendStructure(b, bindings);
			}
		}
		b.append(")");
	}

	public int getIndex() {
		return index;
	}

	public String getName() {
		return name;
	}

	public int getParamCount() {
		return getCardinality();
	}

	public int getCardinality() {
		return nodes == null ? 0 : nodes.size();
	}

	public AstNode getChild(int i) {
		return nodes == null ? null : nodes.get(i);
	}
}
